// $Id: BasePair.h 849 2015-03-27 20:31:19Z ge $
/// \file BasePair.h
/// \brief contains the BasePair class
///
/// $Revision: 849 $
/// \author Gerald Weber

#ifndef GBC_EXP_BASEPAIR_H
#define GBC_EXP_BASEPAIR_H "$Id: BasePair.h 849 2015-03-27 20:31:19Z ge $"

#include "JobControl.h"
#include "Nucleotide.h"

namespace gbc {

template<class _InternalTp=char>
/// \brief Forms a base pair out of two Nucleotides.
/// 
/// This class forms a std::pair from Nucleotide<_InternalTp>.
class BasePair: public std::pair<Nucleotide<_InternalTp>,Nucleotide<_InternalTp> >
  {
  public:
    typedef _InternalTp internal_type;                ///< type of _InternalTp
    typedef Nucleotide<_InternalTp> nucleotide_type;  ///< type of Nucleotide<_InternalTp>
    typedef std::pair<Nucleotide<_InternalTp>,Nucleotide<_InternalTp> > pair_type;
    typedef std::pair<internal_type,internal_type> internal_pair_type;
    typedef std::string version_type;
    
    version_type Version;     ///< Version of base pair eg AT^a
    static const char version_char='^';
    
    enum SymmetryActionsType {do_not_simplify_symmetry = 1 /* formely context_dependent=true */, 
                              simplify_symmetry        = 0 /* formely not_context_dependent=false */};

    static const SymmetryActionsType default_symmetry_action=simplify_symmetry;

    SymmetryActionsType Symmetry_action; ///< Base pairs may simplified according to their 5´->3' directions, for example TA would become AT 
                                         ///< if Symmetry_action=simplify_symmetry
    std::string separation_string;       ///< Separator string for the string representation.

  BasePair(void): Version(), Symmetry_action(simplify_symmetry), separation_string() {}
  
  BasePair(char fst, char scnd, SymmetryActionsType symmetry_action=simplify_symmetry, std::string version=std::string())
  /// \brief Concstructs a BasePair from its named Nucleotides.
  ///
  /// This constructor works by converting the char representation
  /// to a type Nucelotide which is then assigned to the BasePair.
   : pair_type(nucleotide_type(fst),nucleotide_type(scnd)), Version(version),
     Symmetry_action(symmetry_action), separation_string()
   {}

  BasePair(char fst, SymmetryActionsType symmetry_action=simplify_symmetry)
  /// \brief Concstructs a BasePair from its named Nucleotide.
  ///
  /// This constructor works by converting the char representation
  /// to a type Nucelotide which is then assigned to the BasePair.
   : pair_type(nucleotide_type(fst),nucleotide_type(fst)),
     Symmetry_action(symmetry_action), separation_string()
   {
   this->second.complementary();
   }

  BasePair(std::string bpname, SymmetryActionsType symmetry_action=default_symmetry_action)
  /// \brief Concstructs a BasePair from its named Nucleotides.
  ///
  /// This constructor works by converting the string representation
  /// to a type Nucelotide which is then assigned to the BasePair.
   : Symmetry_action(symmetry_action), separation_string()
   {
   std::size_t upper = bpname.find(version_char);
   if (upper != std::string::npos)
     {
     std::string bp=bpname.substr(0,upper);
     this->first=nucleotide_type(bp[0]);
     this->second=nucleotide_type(bp[1]);
     Version=bpname.substr(upper+1);
     }
   else
     {
     this->first=nucleotide_type(bpname[0]);
     this->second=nucleotide_type(bpname[1]);
     }
   }

  template<class _InternalTp2>
  /// \brief Constructs a BasePair from Nucleotide and its complement.
  BasePair(const Nucleotide<_InternalTp2> fst, SymmetryActionsType symmetry_action=default_symmetry_action)
    : pair_type(fst,fst),
      Symmetry_action(symmetry_action), separation_string()
    {
    this->second.complementary();
    }

  template<class _InternalTp2>
  /// \brief Assigns two Nucleotide objects to form a BasePair.
  BasePair(const Nucleotide<_InternalTp2> fst, const Nucleotide<_InternalTp2> scnd, SymmetryActionsType symmetry_action=default_symmetry_action)
    : pair_type(fst,scnd), 
      Symmetry_action(symmetry_action), separation_string()
    {}

  template<class _InternalTp2>
  /// \brief Assigns a nucleotide, and its complement, to the BasePair.
  BasePair& operator=(const Nucleotide<_InternalTp2>& nuc)
    {
    this->first=nuc;
    this->second=nuc;
    this->second.complementary();

    return *this;
    }

  /// \brief Swaps the two Nucleotides of the Base Pairs.
  void swap(void)
    {
    nucleotide_type tmp(this->second);
    this->second=this->first;
    this->first=tmp;
    }

  /// \brief Converts this BasePair to it's cheapest symmetry equivalent.
  ///
  /// We apply the symmetry operations of swapping.
  /// Then we retain the one that yields the lowest internal
  /// representation value using the operator<. 
  ///
  /// \attention No operation is carried out if Symmetry_action is do_not_simplify_symmetry
  inline bool reduce_to_smallest_symmetry(void)
    {
    bool reduced=false;
    if (Symmetry_action==simplify_symmetry)
      { 
      BasePair sw=*this;
      sw.swap();
      if (sw < *this) {*this=sw; reduced=true;}
      }
    return reduced;
    }

  /// \brief Produces a string which describes the BasePair.
  ///
  /// This method makes use of the separation_string which by default is set to a 
  /// empty string. Possible uses are "/" which would give results such as
  /// "A/T" or "C/G". 
  operator std::string(void) const
    {
    std::string res=this->first.symbol()+separation_string+this->second.symbol();
    if (Version != std::string())
      {
      res.append(1,version_char);
      res.append(this->Version);
      }
    return res;
    }
    
  //This solves insertion problems with ReferenceSet
  //Without overloading this operator bp1 and bp2 may return equal even if they have different versions
  friend bool operator<(const BasePair& bp1, const BasePair& bp2)
    {
    pair_type p1=static_cast<pair_type>(bp1), p2=static_cast<pair_type>(bp2);
    if (p1 == p2) return bp1.Version < bp2.Version;
    else          return p1 < p2;
    }

  operator internal_pair_type(void) const
    {return internal_pair_type(this->first.representation,this->second.representation);}
  
  /// \brief Swaps the two Nucleotides of the Base Pairs.
  ///
  /// This method is called axis_swap such that it is later not
  /// confused with stack_swap of the BasePairNeighbours class.
  friend inline BasePair axis_swap(const BasePair& bp)
    {
    BasePair newpair=bp;
    newpair.first=bp.second;
    newpair.second=bp.first;
    return newpair;
    }

  /// \brief Checks if two base pairs are axis-symmetric to each other. 
  /// \attention This is done regardless of the 3'->5' direction by
  /// simply checking the the representations
  friend inline bool axis_symmetry(const BasePair& bp1, const BasePair& bp2)
    {
    return   (bp1.first.representation  == bp2.second.representation)
          && (bp1.second.representation == bp2.first.representation);
    }

  /// \brief Cheks if the two nucleotides are the same.
  friend inline bool check_symmetry(const BasePair& bp)
    {
    return bp.first.representation == bp.second.representation;
    }

//   friend inline BasePair representation(internal_pair_type pr)
//     {
//     Nucleotide<internal_type> N1, N2;
//     N1.representation=pr.first; N2.representation=pr.second;
//     return BasePair(N1,N2);
//     }

  /// Extractor for printing BasePair symbols
  ///
  /// This method first converts the BasePair to its string representation.
  friend std::ostream& operator<<(std::ostream &out,const BasePair &bp)
    { 
    out << static_cast<std::string>(bp);
    return out;
    }
  };//ends class

//Cannot make a friend specialization
//   template<>
//   friend bool operator<(const BasePair<int>& bp1, const BasePair<int>& bp2)
//     {
//     return (std::string)bp1 < (std::string)bp2;
//     }


};//ends namespace
#endif
