// $Id: IndividualNearestNeighbours.h 865 2015-04-21 22:41:49Z ge $
/// \file IndividualNearestNeighbours.h
/// \brief contains the Gibbs Free Energy function class
///
/// $Revision: 865 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_INDIVIDUALNEARESTNEIGHBOURS_H
#define GBC_INDIVIDUALNEARESTNEIGHBOURS_H "$Id: IndividualNearestNeighbours.h 865 2015-04-21 22:41:49Z ge $"
#include "NeighbourSequence.h"
#include "Duplex.h"
#include <valarray>
#include <fstream>
#include "ParameterMap.h"

using namespace gbc;
namespace gbc
  {
  template<class _Tp=double>
  /// \brief Calculate Gibbs Free Energies and Melting Temperatures
  ///
  /// This implements DNA melting temperature calculations based on the independent nearest neighbour model (INN)
  /// See santalucia96, santalucia98, gray97, gray97b
  class IndividualNearestNeighbours
    {
    public:
      typedef _Tp                              value_type;
      typedef ParameterMap<value_type>         parameter_map_type;
      typedef std::set<std::string>            nearest_neighbours_map_type;
      typedef std::map<std::string,value_type> enthalpy_map_type;
      typedef std::valarray<value_type> vector_type;
      typedef SequenceInfo<false,value_type>   sequenceinfo_type;//false refers to periodicity
      typedef typename sequenceinfo_type::neighboursequence_type
                                               neighboursequence_type;
      typedef std::deque<std::string>          string_deque_type;

#if __GNUC__ > 5
    constexpr
#endif
      static const value_type R=1.987;   ///< Gas constant, cal/(K mol)

      sequenceinfo_type*  Sequence_info;
      bool                Debug;
      value_type          Sequence_enthalpy;      ///< Contains the total sequence ethalphy
      value_type          Sequence_entropy;       ///< Contains the total sequence entropy
      value_type          Initiation_CG_entropy;  ///< For sequences that contain at least one CG base-pair
      value_type          Initiation_CG_enthalpy; ///< For sequences that contain at least one CG base-pair
      value_type          Initiation_AT_entropy;  ///< For sequences that are AT only
      value_type          Initiation_AT_enthalpy; ///< For sequences that are AT only
      value_type          Initiation_entropy;     ///< Entropy for initiation
      value_type          Initiation_enthalpy;    ///< Enthalpy for initiation
      value_type          Symmetry_correction_entropy;    ///< Symmetry correction of self-complementary sequences
      value_type          Symmetry_correction_enthalpy;    ///< Symmetry correction of self-complementary sequences
      value_type          Terminal_5TA3_enthalpy; ///< AT 5' terminal correction
      value_type          Terminal_AU_enthalpy;   ///< Contribution for AU or UA at ends
      value_type          Terminal_AU_entropy;    ///< Contribution for AU or UA at ends
      value_type          Terminal_AT_enthalpy;   ///< Contribution for AT or TA at ends
      value_type          Terminal_AT_entropy;    ///< Contribution for AT or TA at ends
      value_type          Terminal_CG_enthalpy;   ///< Contribution for AT or TA at ends
      value_type          Terminal_CG_entropy;    ///< Contribution for AT or TA at ends
      parameter_map_type  parameter_map;          ///< Contains all parameters
      enthalpy_map_type   Enthalpies, Entropies;  ///< Contains all nearest-neighbour enthalpies and entropies

    IndividualNearestNeighbours(void) 
        : Debug(false), 
          Sequence_enthalpy(), Sequence_entropy() {}

    inline void reset_calculation(void)
      {
      Sequence_enthalpy=value_type();
      Sequence_entropy=value_type();
      }

    /// \brief dummy function, needed in ParameterSearch, has no function
    inline void temperature(value_type T) {}

    /// \brief dummy function, needed in ParameterSearch, has no function
    inline value_type temperature(void) {return value_type();}

    /// \brief dummy function, needed in ParameterSearch, has no function
    friend std::ostream& operator<<(std::ostream &out, const IndividualNearestNeighbours& tm)
      {
      out << tm.parameter_map;
      return out;
      }

    inline bool get_parameters_from_file(const std::string Filename)
      {
      parameter_map.read_from(Filename);
       set_TM_parameters();
      return true;
      }
      
    /// \brief Reads a deque of parameter files and assigns to all maps
    /// All parameters are places into parameter_map.
    inline bool get_parameters_from_list_of_files(string_deque_type &par_file_deque)
      {
      parameter_map.read_from_list_of_files(par_file_deque);
      set_TM_parameters();
      return true;
      }


    void  set_TM_parameters(void) 
      {
      nearest_neighbours_map_type nn;
      parameter_map.get_all_id(nn); // Here we get all combinations of NN from the parameter file
      typename nearest_neighbours_map_type::iterator it;
      for(it=nn.begin(); it != nn.end(); ++it)
        {
        if (it->size() > 2)
          {
          std::string parameter_name=*it + ":" + "enthalpy";
          Enthalpies[*it]=parameter_map[parameter_name];
          
          parameter_name=*it + ":" + "entropy";
          Entropies[*it] =parameter_map[parameter_name];
          }
        }
      Initiation_entropy=parameter_map.get_if_exists("initiation_entropy");
        
      Initiation_enthalpy=parameter_map.get_if_exists("initiation_enthalpy");

      Initiation_CG_entropy=parameter_map.get_if_exists("initiation_CG_entropy");
      
      Initiation_CG_enthalpy=parameter_map.get_if_exists("initiation_CG_enthalpy");

      Initiation_AT_entropy=parameter_map.get_if_exists("initiation_AT_entropy");
	
      Initiation_AT_enthalpy=parameter_map.get_if_exists("initiation_AT_enthalpy");

      Symmetry_correction_entropy=parameter_map.get_if_exists("symmetry_correction_entropy");
      
      Symmetry_correction_enthalpy=parameter_map.get_if_exists("symmetry_correction_enthalpy");
      
      Terminal_5TA3_enthalpy=parameter_map.get_if_exists("terminal_5TA3_enthalpy");
      
      Terminal_AU_enthalpy=parameter_map.get_if_exists("terminal_AU_enthalpy");

      Terminal_AU_entropy=parameter_map.get_if_exists("terminal_AU_entropy");

      Terminal_AT_enthalpy=parameter_map.get_if_exists("terminal_AT_enthalpy");

      Terminal_AT_entropy=parameter_map.get_if_exists("terminal_AT_entropy");

      Terminal_CG_enthalpy=parameter_map.get_if_exists("terminal_CG_enthalpy");

      Terminal_CG_entropy=parameter_map.get_if_exists("terminal_CG_entropy");

      if ((Enthalpies.size() == size_t()) || (Entropies.size() == size_t())) 
        std::cout << __FILE__ << ":" << __LINE__ << " no specific Enthalpy or Entropy values, falling back to generic values" << std::endl;
      }

    /// \returns Enthalpy in Kcal/mol
    /// \returns Entropy in cal/(mol.K)
    inline void calculate_total_enthalpy_and_entropy(void)
      {
      value_type total_enthalpy=value_type();
      value_type total_entropy =value_type();
      if (Debug) std::cerr << "Sequence " << Sequence_info->Exact_duplex << std::endl;
      typename neighboursequence_type::const_iterator it;
      for(it=Sequence_info->Reduced_neighbours.begin(); 
          it != Sequence_info->Reduced_neighbours.end(); ++it)
        {
        if (Debug) std::cerr << *it;
        typename enthalpy_map_type::iterator enth=Enthalpies.find(static_cast<std::string>(*it));
        if (enth == Enthalpies.end())
          {
          std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Enthalphy for " << *it << " not found" << std::endl;
          std::cerr << "Enthalphies[] has " << Enthalpies.size() << " elements" << std::endl;
          }
        else total_enthalpy += Enthalpies[static_cast<std::string>(*it)];
        if (Debug) 
          {
          std::cerr << " DH=" <<  Enthalpies[static_cast<std::string>(*it)];
          }
        typename enthalpy_map_type::iterator entr=Entropies.find(static_cast<std::string>(*it));
        if (entr == Entropies.end())
          {
          std::cerr << __FILE__ << " " << __LINE__ << "\n" << "Entropy for " << *it << " not found" << std::endl;
          std::cerr << "Entropies[] has " << Entropies.size() << " elements" << std::endl;
          }
        else total_entropy += Entropies[static_cast<std::string>(*it)];
        if (Debug) 
          {
          std::cerr << " DS =" <<  Entropies[static_cast<std::string>(*it)] << std::endl;
          }
        }
      if (Sequence_info->Has_CG)    
        {
        total_enthalpy += Initiation_CG_enthalpy; 
        if (Debug && (Initiation_CG_enthalpy != 0)) std::cerr << " Initiation CG enthalpy= " << Initiation_CG_enthalpy << std::endl;
        total_entropy  += Initiation_CG_entropy; 
        if (Debug && (Initiation_CG_entropy != 0))  std::cerr << " Initiation CG entropy= " << Initiation_CG_entropy << std::endl;
        }  
      if (Sequence_info->Is_AT_only)   
        {
        total_enthalpy += Initiation_AT_enthalpy;
        if (Debug && (Initiation_AT_enthalpy != 0)) std::cerr << " Initiation AT enthalpy = " << Initiation_AT_enthalpy << std::endl;
        total_entropy += Initiation_AT_entropy;
        if (Debug && (Initiation_AT_entropy != 0))  std::cerr << " Initiation AT entropy = " << Initiation_AT_entropy << std::endl;
        }
      if (Sequence_info->Terminal_5TA3)
        {
        total_enthalpy += Sequence_info->Terminal_5TA3*Terminal_5TA3_enthalpy;
        if (Debug && (Terminal_5TA3_enthalpy != 0)) 
          std::cerr << "Terminal_5TA3_enthalpy = " << Sequence_info->Terminal_5TA3 << " X " << Terminal_5TA3_enthalpy << " = " << Sequence_info->Terminal_5TA3*Terminal_5TA3_enthalpy << std::endl;
        }
      if (Sequence_info->Terminal_AU)
        {
        total_enthalpy += Sequence_info->Terminal_AU*Terminal_AU_enthalpy;
        total_entropy += Sequence_info->Terminal_AU*Terminal_AU_entropy;
        if (Debug && (Terminal_AU_enthalpy != 0)) 
          std::cerr << "Terminal_AU_enthalpy = " << Sequence_info->Terminal_AU << " X " << Terminal_AU_enthalpy << " = " << Sequence_info->Terminal_AU*Terminal_AU_enthalpy << std::endl;
        if (Debug && (Terminal_AU_entropy != 0))  
          std::cerr << "Terminal_AU_entropy = " << Sequence_info->Terminal_AU << " X " << Terminal_AU_entropy << " = "  << Sequence_info->Terminal_AU*Terminal_AU_entropy << std::endl;
        }
      if (Sequence_info->Terminal_AT)
        {
        total_enthalpy += Sequence_info->Terminal_AT*Terminal_AT_enthalpy;
        total_entropy += Sequence_info->Terminal_AT*Terminal_AT_entropy;
        if (Debug && (Terminal_AT_enthalpy != 0)) 
          std::cerr << "Terminal_AT_enthalpy = " << Sequence_info->Terminal_AT << " X " << Terminal_AT_enthalpy << " = " << Sequence_info->Terminal_AT*Terminal_AT_enthalpy << std::endl;
        if (Debug && (Terminal_AT_entropy != 0))  
          std::cerr << "Terminal_AT_entropy = " << Sequence_info->Terminal_AT << " X " << Terminal_AT_entropy << " = "  << Sequence_info->Terminal_AT*Terminal_AT_entropy << std::endl;
        }
      if (Sequence_info->Terminal_CG)
        {
        total_enthalpy += Sequence_info->Terminal_CG*Terminal_CG_enthalpy;
        total_entropy  += Sequence_info->Terminal_CG*Terminal_CG_entropy;
        if (Debug && (Terminal_CG_enthalpy != 0)) 
          std::cerr << "Terminal_CG_enthalpy = " << Sequence_info->Terminal_CG << " X " << Terminal_CG_enthalpy << " = " << Sequence_info->Terminal_CG*Terminal_CG_enthalpy << std::endl;
        if (Debug && (Terminal_CG_entropy != 0))  
          std::cerr << "Terminal_CG_entropy = " << Sequence_info->Terminal_CG << " X " << Terminal_CG_entropy << " = "  << Sequence_info->Terminal_CG*Terminal_CG_entropy << std::endl;
        }

      if (Initiation_enthalpy != 0)
        {
        total_enthalpy += Initiation_enthalpy;
        if (Debug)  std::cerr << "Initiation_enthalpy = " << Initiation_enthalpy << std::endl;
        }
      if (Initiation_entropy != 0)
        {
        total_entropy += Initiation_entropy;
        if (Debug)  std::cerr << "Initiation_entropy = " << Initiation_entropy << std::endl;
        }
        
      if (Sequence_info->Self_complementary) 
        {
        total_entropy += Symmetry_correction_entropy;
        total_enthalpy += Symmetry_correction_enthalpy;
        if (Debug) std::cerr << "Symmetry correction entropy = " << Symmetry_correction_entropy << std::endl;
        if (Debug) std::cerr << "Symmetry correction enthalpy = " << Symmetry_correction_enthalpy << std::endl;
        }
      
      Sequence_enthalpy=Sequence_info->enthalpy.predicted=total_enthalpy;
      Sequence_entropy=Sequence_info->entropy.predicted=total_entropy;
      if (Debug)  std::cerr << "Total DH = " << Sequence_enthalpy << " Total DS = " << Sequence_entropy << std::endl;
      }

    /// \returns Enthalpy in Kcal/mol
    inline value_type total_enthalpy(void)
      {
      if (Sequence_enthalpy == value_type()) calculate_total_enthalpy_and_entropy();
      return Sequence_enthalpy;
      }

    /// \returns Entropy in cal/(mol.K)
    inline value_type total_entropy(void)
      {
      if (Sequence_entropy == value_type()) calculate_total_enthalpy_and_entropy();
      return Sequence_entropy;
      }

    inline value_type melting_temperature(void)
      {
      value_type alpha=4.0;
      if (Sequence_info->Self_complementary) alpha=1.0;
      if (Sequence_info->species_concentration == value_type())
        std::cerr << __FILE__ << ":" << __LINE__ << " warning: species concentration is zero (Ct=0), temperature calculation will fail due to ln(0) calculation " << std::endl;
      Sequence_info->temperature.predicted=total_enthalpy()*1000.0/(total_entropy()+R*log(Sequence_info->species_concentration*sequenceinfo_type::Concentration_factor/alpha))-273.15;
      return Sequence_info->temperature.predicted;
      }

    ///< \brief Calculates the Gibbs Free Energy (in kcal/mol)
    ///
    /// If the temperature is not given we assume 37 C.
    /// If entropy and ethalpy are not given, we calculate them.
    inline value_type gibbs_free_energy(value_type temp=37.0,
                                        value_type enth=value_type(),
                                        value_type entr=value_type())
      {
      if (enth == value_type()) enth=total_enthalpy();
      if (entr == value_type()) entr=total_entropy();
      return (enth*1000.0-(273.15+temp)*entr)/1000.0;
      }

    };

  };

#endif
