// $Id: SequenceDataset.h 883 2015-07-06 20:41:09Z ge $
/// \file SequenceDataset.h
/// \brief This class manages datasest of experimental sequence information
///
/// $Revision: 883 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_SEQUENCEDATASET_H
#define GBC_SEQUENCEDATASET_H "$Id: SequenceDataset.h 883 2015-07-06 20:41:09Z ge $"

#include "SequenceInfo.h"
#include <deque>
#include <vector>
#include "MathAux.h"
#include<boost/algorithm/string/split.hpp>                                      
#include<boost/algorithm/string.hpp>                                            

namespace gbc
{
template<class _Tp=double>
class DataStatistics
  {
  typedef _Tp value_type;
  public:
    std::string name;
    value_type error;
    value_type average;
    value_type diff_deviation;
    value_type std_deviation;
    value_type sqr_diff;
    value_type sqrt_diff2;

  DataStatistics(void)
    : name(), error(), average(), diff_deviation(), 
      std_deviation(), sqr_diff(), sqrt_diff2() {}

  DataStatistics(std::string nm)
    : name(nm), error(), average(), diff_deviation(), 
      std_deviation(), sqr_diff(), sqrt_diff2() {}

  inline void print_name(std::ostream &out) const
    {
    out << name << ".error " 
        << name << ".average "
        << name << ".std_deviation "
        << name << ".sqr_diff"; 
    }

  inline friend std::ostream& operator<<(std::ostream &out, DataStatistics& dt)
    {
    out << dt.error << " " << dt.average << " " << dt.std_deviation << " " << dt.sqr_diff;
    return out;
    }
  };

template<class _SequenceInfo>
class SequenceDataset
  {
  public:
    typedef _SequenceInfo                        sequence_info_type;
    typedef typename sequence_info_type::sequence_type
                                                 sequence_type;
    typedef size_t                               sequence_length_type;
    typedef typename sequence_info_type::value_type   
                                                 value_type;
    typedef typename sequence_info_type::equivalence_map_type    equivalence_map_type;
    typedef typename sequence_info_type::duplex_type    duplex_type;
    typedef value_type                           salt_concentration_type;
    typedef std::set<salt_concentration_type>    salt_set_type;
    typedef std::set<value_type>                 length_set_type;
    typedef std::deque<sequence_info_type>       sequence_info_deque_type;
    typedef std::deque<sequence_info_type*>      sequence_info_deque_ptype;
    typedef std::map<sequence_length_type,
                     sequence_info_deque_ptype>  length_dataset_type;
    typedef std::map<salt_concentration_type,
                     length_dataset_type>        salt_length_dataset_type;
    typedef std::deque<value_type>               species_concentration_deque_type;
    typedef typename salt_length_dataset_type::iterator 
                                                 iterator;
    typedef std::valarray<value_type>            vector_type;
    typedef DataStatistics<value_type>           datastatistics_type;
    typedef ReferenceSet<Duplex<> >                    duplex_ref_type;
    typedef ReferenceSet<NeighbourSequence<> >         neighbours_ref_type;
    typedef ReferenceSet<StrandPairSequence<> >        strandpair_ref_type;
    typedef std::deque<std::string>                    string_deque_type;

  std::string Dataset_identfifier;
  std::string Dataset_type;
  bool Uniform_species_concentration;
  bool Species_concentration_necessary;
  value_type Species_concentration_symmetry_factor;       ///< default = 1.0
  species_concentration_deque_type Species_concentration;
  bool Uniform_salt_concentration;
  value_type Salt_concentration;
  std::string Dataset_filename;
  sequence_info_deque_type  Raw_dataset;
  sequence_info_deque_type  Original_raw_dataset;
  salt_set_type             Salt_concentrations;          ///< Set of salt concentrations
  length_set_type           Lengths;                      ///< Set of lenght values

  length_dataset_type       Length_dataset;               ///< A map indexed by length
  salt_length_dataset_type  Salt_length_dataset;          ///< a map index by salt and length
  salt_length_dataset_type  Original_salt_length_dataset;

  datastatistics_type Temperature;
  datastatistics_type Entropy;
  datastatistics_type Enthalpy;

  std::deque<value_type> diff;

  int prediction_method;
  
  equivalence_map_type EquivalenceMap; ///< Map which collects base-pair equivalences 
  duplex_ref_type        Basepair_set;              ///< The set of different base pairs
  neighbours_ref_type    Neighbours_set;          ///< The set of different base pair neighbours
  strandpair_ref_type    StrandPair_set;          ///< The set of different strand pair neighbours

  neighbours_ref_type    NPNeighbours_set;          ///< The set of different base pair neighbours, non-periodic
  strandpair_ref_type    NPStrandPair_set;          ///< The set of different strand pair neighbours, non-periodic


  SequenceDataset(const std::string id=std::string()): Dataset_identfifier(id), Dataset_type(),
    Uniform_species_concentration(true), Species_concentration_necessary(true), Species_concentration_symmetry_factor(1.0), Species_concentration(),
    Uniform_salt_concentration(true), Salt_concentration() {};
  
   // This function collects the different salt concentrations in the dataset, and for each makes
   // a list of available sequence lengths. This is used for the regression method of the
   // Peyrard-Bishop thermodynamic equivalence.
   void make_salt_length_map(void)
      {
      Salt_length_dataset.clear();
      typename sequence_info_deque_type::iterator si=Raw_dataset.begin();
      for(si=Raw_dataset.begin(); si != Raw_dataset.end(); si++)
        { 
        if (Salt_length_dataset.find(si->salt_concentration) == Salt_length_dataset.end())
          {
          Salt_concentrations.insert(si->salt_concentration);
          Salt_length_dataset.insert(typename salt_length_dataset_type::value_type(si->salt_concentration,length_dataset_type()));
          }

        if (Salt_length_dataset[si->salt_concentration].find(si->length) == Salt_length_dataset[si->salt_concentration].end())
          {
          Lengths.insert(si->length);
          Salt_length_dataset[si->salt_concentration].insert(typename length_dataset_type::value_type(si->length,sequence_info_deque_ptype()));
          }
      
        // Populate the map.
        Salt_length_dataset[si->salt_concentration][si->length].push_back(&(*si));
        }
      }

    sequence_info_type add(const std::string& seq,        ///< The sequence string 
             const std::string& comp,       ///< The complimentary sequence string 
             const value_type& salt_conc,   ///< The salt concentration
             const value_type& temperature, ///< The experimental melting temperature
             const value_type& enthalpy,    ///< The experimental enthalpy
             const value_type& entropy,     ///< The experimental entropy
             const value_type& spec_conc)   ///< Species concentration.
      {
      // If there is no reference species concentration yet, then use the current one.
      if (Species_concentration.empty()) Species_concentration.push_back(spec_conc);
      // Check if species concentration is uniform for dataset
      if (Uniform_species_concentration)
        {
        if (spec_conc != Species_concentration[0]) {Uniform_species_concentration=false;}
        }

      // If there is no reference salt concentration yet, then use the current one.
      if (Salt_concentration == value_type()) Salt_concentration=salt_conc;
      // Check if species concentration is uniform for dataset
      if (Uniform_salt_concentration)
        {
        if (salt_conc != Salt_concentration) {Uniform_salt_concentration=false;}
        }

      sequence_info_type si;
      si.pEquivalenceMap=&EquivalenceMap;//get hold of the trimer liste equivalence map from read_xml
      si.length=seq.length();


      si.salt_concentration=salt_conc;
      si.sequence=sequence_type(seq,comp); //this will become obsolete
      si.insert_sequence(seq,comp);
      Basepair_set.insert_from_map(si.Basepair_set);
      Neighbours_set.insert_from_map(si.Neighbours_set);
      StrandPair_set.insert_from_map(si.StrandPair_set);
      NPNeighbours_set.insert_from_map(si.NPNeighbours_set);
      NPStrandPair_set.insert_from_map(si.NPStrandPair_set);
      si.determine_alpha();
      
      if (spec_conc == value_type()) 
        {
        if ((si.species_concentration == value_type()) and not Species_concentration.empty()) si.species_concentration=Species_concentration[0];
        }
      else si.species_concentration=spec_conc;

      if (si.Self_complementary) si.species_concentration *= Species_concentration_symmetry_factor;

      
      si.temperature.adjusted=si.temperature.measured=temperature;
      si.enthalpy.measured=enthalpy;
      si.entropy.measured=entropy;
      if (not Species_concentration.empty()) si.Ref_concentration=Species_concentration[0];//Ref_concentration no in use

      Raw_dataset.push_back(si);
      return si;
      }
      
    /// Update
    void update(sequence_info_type &si)
      {
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
	if (si.sequence == sequence_info_it->sequence)
	  {
	  *sequence_info_it = si;
	  }
	}
      }


    /// \brief Adds a new sequence and its experimental data to a map.
    ///
    /// This is one of the main functions of this class. It adds a new sequence
    /// and its experimental melting temperature, together with salt and species
    /// concentration to a map. 
    sequence_info_type add_temperature(const std::string& seq,        ///< The sequence string 
             const std::string& comp,       ///< The complementary sequence string 
             const value_type& salt_conc,   ///< The salt concentration in mmol/L
             const value_type& temperature, ///< The experimental melting temperature in K
             const value_type& spec_conc)   ///< Species concentration in 1e-6 M
      {
      if (spec_conc == value_type() and Species_concentration_necessary)
        std::cerr << __FILE__ << ":" << __LINE__ << " warning: species concentration is zero (Ct=0) " << std::endl;
      add(seq,comp,salt_conc,temperature,0.0,0.0,spec_conc);
      sequence_info_type *si = &(Raw_dataset.back()); //gets a pointer to itself from the Raw_dataset map
      return *si;
      }

    /// \brief Adds a new sequence and its experimental data to a map.
    ///
    /// This is one of the main functions of this class. It adds a new sequence
    /// and its experimental enthalpy/entropy, together with salt and species
    /// concentration to a map. 
    sequence_info_type add_gibbs(const std::string& seq,        ///< The sequence string 
             const std::string& comp,       ///< The complementary sequence string 
             const value_type& salt_conc,   ///< The salt concentration
             const value_type& enthalpy,    ///< The experimental enthalpy
             const value_type& entropy,     ///< The experimental entropy
             const value_type& spec_conc)   ///< Species concentration.
      {
      if (spec_conc == value_type())
        std::cerr << __FILE__ << ":" << __LINE__ << " warning: species concentration is zero (Ct=0) " << std::endl;

      add(seq,comp,salt_conc,0.0,enthalpy,entropy,spec_conc);
      sequence_info_type *si = &(Raw_dataset.back());  //gets a pointer to itself from the Raw_dataset map
      if (spec_conc != value_type())
        {
        si->set_temperature_from_gibbs();
        }
      return *si;
      }
      
    /// Despite its name no xml is read, its an ordinary data file organized in columns  
    void read_xml(std::string xmlfile, species_concentration_deque_type spconc=species_concentration_deque_type(), double multspconc=1)
      {
      std::ifstream xml;
      xml.open(xmlfile.c_str());
      if (!xml.is_open())
        {
        std::cout << __FILE__ << ":" << __LINE__ << " Could not read " << xmlfile << std::endl;
        exit(1);
        }
      std::string type;
      xml >> type;
      std::string seq, comp;
      double salt, concentration;
      if (spconc.empty()) spconc=Species_concentration;
      if (type == std::string("gibbs"))
        {
        Dataset_type="gibbs";
        double enthalpy, entropy;
        while(!xml.eof())
          {
          xml >> seq >> comp 
              >> enthalpy        //in kcal/mol
              >> entropy         //in cal/(mol.K) or eu
              >> salt;           //in mM or mmol/L
          if (!xml.eof())
            {
            if (spconc.size() > 0)
              {
              // Here we add the sequence for each Ct in the list
              typename species_concentration_deque_type::iterator ctlist;
              for (ctlist=spconc.begin(); ctlist != spconc.end(); ++ctlist)
                add_gibbs(seq,comp,salt,enthalpy,entropy,(*ctlist)*multspconc);
              }
            else
              {
              //adding with zero species_concentration
              add_gibbs(seq,comp,salt,enthalpy,entropy,0.0);
              }
            }
          }
        }      
      else
        {
        if (type == std::string("temperature"))
          {
          Dataset_type="temperature";
          double temp;
          while(!xml.eof())
            {
            xml >> seq >> comp 
                >> temp 
                >> salt            //in C
                >> concentration; //\mu M
            if (not spconc.empty()) concentration = spconc[0]; //replace species concentration of the file
            if (multspconc != 1.0) concentration *= multspconc; //multiply species concentration of the file
            if (!xml.eof())
              {
              add_temperature(seq,comp,salt,temp,concentration);
              }
            }
          }
        else
          {
          std::cerr << "Format of data file " << xmlfile.c_str() << " not recongnized, make sure first line is either \'temperature\' or \'gibbs\'" << std::endl;
          exit(1);
          }
        }
      make_salt_length_map();
      Original_raw_dataset=Raw_dataset;
      }
      
    void read_rules(std::string rulefile)
      {
      std::ifstream rules;
      rules.open(rulefile.c_str());
      if (!rules.is_open())
        {
        std::cerr << __FILE__ << ":" << __LINE__ << " Could not read " << rulefile << std::endl;
        exit(1);
        }

      while(!rules.eof())
        {
        std::string line;
        std::getline(rules,line);
        if (line[0] == '+')
          {
          line.erase(0,1);
          std::deque<std::string> line_split;
          boost::algorithm::split(line_split, line, boost::is_any_of(" "),boost::token_compress_on);
          if (line_split.size() != 2) 
            std::cerr << __FILE__ << ":" << __LINE__ << "Could not interpret rule line [" << line << "] check for formatting mistakes" << std::endl;
          std::string bp=line_split[0];
          std::string equiv=line_split[1];;
          std::deque<std::string> extracted;
          boost::algorithm::split(extracted, equiv, boost::is_any_of(","),boost::token_compress_on);
          std::deque<std::string>::iterator ei;
          for(ei=extracted.begin(); ei != extracted.end(); ei++)
            {
            if (ei->length() < 7) std::cerr << __FILE__ << ":" << __LINE__ << " Possibly malformed trimer " << *ei << std::endl;
            duplex_type tmp(*ei);
            tmp.reduce_to_smallest_duplex_symmetry();
            EquivalenceMap[tmp]=bp;
            }
          }
        }
      typename equivalence_map_type::iterator eq=EquivalenceMap.begin();
      if (eq != EquivalenceMap.end())
        {
        std::cout << std::endl << "Equivalences read from file " << rulefile << std::endl;
        }
      for(eq=EquivalenceMap.begin(); eq != EquivalenceMap.end(); eq++)
        {
        std::cout << (std::string)eq->first << " <=> " << eq->second << std::endl;
        }
      }
      
    void read_rules_from_list_of_files(string_deque_type &par_file_deque)
      {
      typename string_deque_type::iterator pit;
      for(pit=par_file_deque.begin(); pit != par_file_deque.end(); pit++)
        this->read_rules(*pit);
      }


    void randomise_dataset(void)
      {
      Raw_dataset=Original_raw_dataset;
      make_salt_length_map();

      std::vector<value_type> temperature_change,enthalpy_change,entropy_change;
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
         if (Dataset_type == std::string("temperature"))
           {
           temperature_change.push_back(gauss_ran(Temperature.error));
           sequence_info_it->temperature.measured += temperature_change.back();
           sequence_info_it->temperature.adjusted=sequence_info_it->temperature.measured;
           }
         if (Dataset_type == std::string("gibbs"))
           {
           enthalpy_change.push_back(gauss_ran(sequence_info_it->enthalpy.measured*Enthalpy.error*0.01));
           entropy_change.push_back(gauss_ran(sequence_info_it->entropy.measured*Entropy.error*0.01));
           sequence_info_it->enthalpy.measured += enthalpy_change.back();
           sequence_info_it->entropy.measured += entropy_change.back() ;
           value_type tmpch=sequence_info_it->temperature.measured;
           sequence_info_it->set_temperature_from_gibbs();
           temperature_change.push_back(tmpch-sequence_info_it->temperature.measured);
           }
         }
      Temperature.std_deviation=standard_deviation(temperature_change);
      Enthalpy.std_deviation=standard_deviation(enthalpy_change);
      Entropy.std_deviation=standard_deviation(entropy_change);
      }



    /// \brief Calculates the difference between the predicted and the adjusted temperature
    inline void difference(void)
      {
      diff.clear();
      value_type sum2=value_type();
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
        value_type delta=fabs(sequence_info_it->temperature.predicted-sequence_info_it->temperature.adjusted);
        diff.push_back(delta);
        sum2 += pow(delta,2);
        }
      Temperature.average=average(diff);
      Temperature.diff_deviation=standard_deviation(diff);
      Temperature.sqr_diff=sqr_average(diff);
      Temperature.sqrt_diff2=sqrt(sum2/Raw_dataset.size());
      }

    /// \brief Prints out the experimental and calculated temperatures.
    inline void print_temperatures(std::ostream& out, bool analyse=false)
      {
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      sequence_info_it->print_head(out);
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
        sequence_info_it->print(out,analyse);
        }
      if (analyse)
        {
        out << this->Basepair_set.size() << " unique base-pairs in dataset (symmetry reduced)" <<  std::endl;
        out << "Base-pair set      : " << this->Basepair_set << std::endl;
        out << this->Neighbours_set.size() << " unique nearest-neighbours in dataset  (symmetry reduced)" <<  std::endl;
        out << "Neighbours set     : " << this->Neighbours_set << std::endl;
        if (this->NPNeighbours_set.size() > 0) out << "Neighbours set (NP): " << this->NPNeighbours_set << std::endl;
        out << this->StrandPair_set.size() << " unique strand-pair-neighbours in dataset (5'->3' direction)" <<  std::endl;
        out << "StrandPair set     : " << this->StrandPair_set << std::endl;
        if (this->NPStrandPair_set.size()) out << "StrandPair set (NP): " << this->NPStrandPair_set << std::endl;
        out << "NP=non-periodic" << std::endl;
        }
      else
        {
        out << "StdTempDiff "     << Temperature.std_deviation << std::endl;
        out << "StdEnthalpyDiff " << Enthalpy.std_deviation    << std::endl;
        out << "StdEntropyDiff "  << Entropy.std_deviation     << std::endl;
        }

      }

   inline void verify(std::ostream& out)
      {
      difference();
      out << Temperature.average << " " << Temperature.diff_deviation << " " << Temperature.sqr_diff << " " << Temperature.sqrt_diff2;
      }

   inline void print_tex(std::ostream& out)
     {
     out << "\\SqrDiff{" << Temperature.sqr_diff   << "}" << std::endl;
     out << "\\AvDiff{"  << Temperature.average   << "}" << std::endl;
     out << "\\StdDiff{" << Temperature.diff_deviation << "}" << std::endl;
     out << "\\StdTempDiff{" << Temperature.std_deviation << "}" << std::endl;
     out << "\\StdEnthalpyDiff{" << Enthalpy.std_deviation << "}" << std::endl;
     out << "\\StdEntropyDiff{" << Entropy.std_deviation << "}" << std::endl;
     }

    /// \brief Build a vector of lengths for a given salt concentration
    inline vector_type length_vector(typename salt_length_dataset_type::iterator& salt_it)
      {
      vector_type lv(salt_it->second.size());
      typename length_dataset_type::iterator length_it;
      size_t n;
      for (length_it=salt_it->second.begin(),n=0; length_it != salt_it->second.end(); ++length_it, ++n)
        {
        lv[n]=length_it->first;
        }
      return lv;
      }

    /// \brief Build a vector of thermal_equivalences for a given length
    inline vector_type thermal_equivalence_vector(typename length_dataset_type::iterator& length_it)
      {
      vector_type lv(length_it->second.size());
      typename sequence_info_deque_ptype::iterator si_it;
      size_t n;
      for (si_it=length_it->second.begin(),n=0; si_it != length_it->second.end(); ++si_it, ++n)
        {
        lv[n]=(*si_it)->thermal_equivalence;
        //std::cerr << "theq=" << lv[n] << std::endl;
        }
      return lv;
      }

    /// \brief Build a vector of all thermal_equivalences
    inline vector_type thermal_equivalence_vector(void)
      {
      vector_type lv(Raw_dataset.size());
      typename sequence_info_deque_type::iterator si=Raw_dataset.begin();
      size_t n;
      for(si=Raw_dataset.begin(),n=0; si != Raw_dataset.end(); si++, ++n)
        {
	//std::cerr << "theq=" << si->thermal_equivalence << std::endl;
        lv[n]=si->thermal_equivalence;
        }
      return lv;
      }

   /// \brief Build a vector of temperatures for a given length
   inline  vector_type temperature_vector(typename length_dataset_type::iterator& length_it)
      {
      vector_type lv(length_it->second.size());
      typename sequence_info_deque_ptype::iterator si_it;
      size_t n;
      for (si_it=length_it->second.begin(),n=0; si_it != length_it->second.end(); ++si_it, ++n)
        {
        lv[n]=(*si_it)->temperature.adjusted;
        //std::cerr << "temp=" << lv[n] << std::endl;
        }
      return lv;
      }

   /// \brief Build a vector of all temperatures
   inline  vector_type temperature_vector(void)
      {
      vector_type lv(Raw_dataset.size());
      typename sequence_info_deque_type::iterator si=Raw_dataset.begin();
      size_t n;
      for(si=Raw_dataset.begin(),n=0; si != Raw_dataset.end(); si++, ++n)
        {
        lv[n]=si->temperature.adjusted;
        //std::cerr << "temp=" << lv[n] << std::endl;
        }
      return lv;
      }
  };
};
#endif